#pragma once


#include <windows.h>
#include <psapi.h>

#include <vector>
#include <filesystem>
#include <chrono>
#include <thread>

#include "../except.hpp"
#include "../string.hpp"
#include "../log.hpp"
#include "../process.hpp"

namespace fcf {
namespace details {

  inline pid_t create_process(const char* a_name,
                              std::vector<std::string> a_argv,
                              std::vector<std::pair<std::string, std::string> > a_envv,
                              const char* a_user,
                              const char* a_password) {
    a_name      = a_name      ? a_name : "";
    a_user      = a_user      ? a_user : "";
    a_password  = a_password  ? a_password : "";

    std::wstring name(fcf::utf8str_to_wstr(a_name));
    std::wstring cmdline;
    if (!name.empty())
      cmdline += std::wstring() + L"\"" + name + L"\"";
    std::string cmdline_cstr;
    if (!name.empty())
      cmdline_cstr += std::string() + "\"" + a_name + "\"";
    bool first = true;
    for (std::string& arg : a_argv) {
      if (!first || (first && !name.empty()))
        cmdline += L" ";
      cmdline += fcf::utf8str_to_wstr(arg);
      if (!first || (first && !name.empty()))
        cmdline_cstr += " ";
      cmdline_cstr += arg;
      first = false;
    }
    std::vector<wchar_t> env;
    for (auto& pair : a_envv) {
      std::wstring key(fcf::utf8str_to_wstr(pair.first));
      std::wstring val(fcf::utf8str_to_wstr(pair.second));
      for (auto& c : key) env.push_back(c);
      env.push_back(L'=');
      for (auto& c : val) env.push_back(c);
      env.push_back(0);
    }
    for (size_t i = 0; environ[i]; ++i) {
      for (size_t j = 0; environ[i][j]; ++j)
        env.push_back(environ[i][j]);
      env.push_back(0);
    }
    for (auto& pair : a_envv) {
      std::wstring key(fcf::utf8str_to_wstr(pair.first));
      std::wstring val(fcf::utf8str_to_wstr(pair.second));
      for (auto& c : key) env.push_back(c);
      env.push_back(L'=');
      for (auto& c : val) env.push_back(c);
      env.push_back(0);
    }
    env.push_back(0);
    env.push_back(0);
    std::wstring user(fcf::utf8str_to_wstr(a_user));
    std::wstring password(fcf::utf8str_to_wstr(a_password));

    SECURITY_ATTRIBUTES sa;
    memset(&sa, 0, sizeof(sa));
    sa.nLength = sizeof(SECURITY_ATTRIBUTES);
    sa.bInheritHandle = TRUE;
    STARTUPINFOW si;
    memset(&si, 0, sizeof(si));
    si.cb = sizeof(STARTUPINFOW);
    si.dwFlags = 0;
    si.hStdOutput = GetStdHandle(STD_OUTPUT_HANDLE);
    si.hStdError = GetStdHandle(STD_ERROR_HANDLE);
    PROCESS_INFORMATION pi;
    memset(&pi, 0, sizeof(pi));

    HANDLE token = 0;
    if (!user.empty()) {
      if (!LogonUserW(user.c_str(), 0, password.c_str(), LOGON32_LOGON_BATCH, LOGON32_PROVIDER_DEFAULT, &token))
        throw fcf::except::authuser(a_user);
      CreateProcessAsUserW(
        token,
        (LPWSTR)name.c_str(),
        (LPWSTR)cmdline.c_str(),
        &sa,
        0,
        FALSE,
        CREATE_UNICODE_ENVIRONMENT,
        &env[0],
        (LPWSTR)std::filesystem::current_path().wstring().c_str(),
        &si,
        &pi
      );
    }
    else {
      CreateProcessW(
        !name.empty() ? (LPWSTR)name.c_str() : (LPWSTR)0,
        (LPWSTR)cmdline.c_str(),
        &sa,
        0,
        FALSE,
        CREATE_UNICODE_ENVIRONMENT,
        &env[0],
        (LPWSTR)std::filesystem::current_path().wstring().c_str(),
        &si,
        &pi
      );
    }

    if (!pi.dwProcessId)
      throw fcf::except::process_start(cmdline_cstr.c_str());

    std::this_thread::sleep_for(std::chrono::seconds(1));

    if (pi.hThread)
      CloseHandle(pi.hThread);
    if (pi.hProcess)
      CloseHandle(pi.hProcess);

    return (pid_t)pi.dwProcessId;
  }

  inline void close_process(pid_t a_pid) {
    HANDLE handle = OpenProcess(PROCESS_TERMINATE, false, a_pid);
    if (handle) {
      TerminateProcess(handle, 1);
      CloseHandle(handle);
    }
  }

  inline size_t get_process_memory_size(int a_pid) {
    HANDLE h = OpenProcess(PROCESS_QUERY_INFORMATION|PROCESS_QUERY_LIMITED_INFORMATION|PROCESS_VM_READ, false, a_pid);
    if (!h)
      return 0;
    PROCESS_MEMORY_COUNTERS info;
    memset(&info, 0, sizeof(info));
    GetProcessMemoryInfo(h, &info, sizeof(info));
    CloseHandle(h);
    return (size_t)info.WorkingSetSize;
    
  }

  inline bool check_pid(pid_t a_pid) {
    HANDLE h = OpenProcess(PROCESS_QUERY_INFORMATION, false, a_pid);
    if (!h)
      return false;
    DWORD exit_code = 0;
    BOOL ret = GetExitCodeProcess(h, &exit_code);
    CloseHandle(h);
    return ret && exit_code == STILL_ACTIVE;
  }



}
}
